from typing import Any, List, Sequence, Tuple, Union

DictIterable = Sequence[Tuple[str, Any]]
JsonStr = str


class CWechatRobotABC:
    """Com Wechat实现的接口"""

    def CStartRobotService(self, wx_pid: int) -> int:
        ...

    def CStopRobotService(self, wx_pid: int) -> int:
        ...

    def CSendImage(self, wx_pid: int, wxid: str, imgpath: str) -> int:
        ...

    def CSendText(self, wx_pid: int, wxid: str, text: str) -> int:
        ...

    def CSendFile(self, wx_pid: int, wxid: str, filepath: str) -> int:
        ...

    def CSendArticle(
        self, wx_pid: int, wxid: str, title: str, abstract: str, url: str, imgpath: str
    ) -> int:
        ...

    def CSendCard(self, wx_pid: int, wxid: str, shared_wxid: str, nickname: str) -> int:
        ...

    def CSendAtText(
        self,
        wx_pid: int,
        chatroomid: str,
        at_wxids: Union[str, List[str]],
        text: str,
        auto_nickname: bool = True,
    ) -> int:
        ...

    def CGetFriendList(
        self,
        wx_pid: int,
    ) -> DictIterable:
        ...

    def CGetFriendListString(self, wx_pid: int):
        ...

    def CGetWxUserInfo(self, wx_pid: int, wxid: str) -> str:  # json
        ...

    def CGetSelfInfo(self, wx_pid: int) -> str:  # json
        ...

    def CCheckFriendStatus(self, wx_pid: int, wxid: str):
        """当前测试不是无痕检测"""
        ...

    def CGetComWorkPath(self, wx_pid: int) -> str:
        ...

    def CStartReceiveMessage(self, wx_pid: int, port: int) -> int:
        """向指定TCP端口发送接收到的消息, 消息结构参考`ReceiveMsgStruct`"""
        ...

    def CStopReceiveMessage(self, wx_pid: int) -> int:
        ...

    def CGetChatRoomMembers(self, wx_pid: int, chatroomid: str) -> DictIterable:
        ...

    def CGetDbHandles(self, wx_pid: int) -> DictIterable:
        ...

    def CExecuteSQL(self, wx_pid: int, handle: int, sql: str) -> list:
        ...

    def CBackupSQLiteDB(self, wx_pid: int, handle: int, savepath: str) -> int:
        ...

    def CVerifyFriendApply(self, wx_pid: int, v3: str, v4: str) -> int:
        ...

    def CAddFriendByWxid(self, wx_pid: int, wxid: str, message: str) -> int:
        ...

    def CAddFriendByV3(self, wx_pid: int, v3: str, message: str) -> int:
        ...

    def CGetWeChatVer(self) -> str:
        ...

    def CStartWeChat(self) -> int:
        ...

    def CSearchContactByNet(self, wx_pid: int, keyword: str) -> DictIterable:
        ...

    def CAddBrandContact(self, wx_pid: int, public_id: str) -> int:
        """添加公众号"""
        ...

    def CHookVoiceMsg(self, wx_pid: int, savepath: str) -> int:
        ...

    def CUnHookVoiceMsg(self, wx_pid: int) -> int:
        ...

    def CHookImageMsg(self, wx_pid: int, savepath: str) -> int:
        ...

    def CUnHookImageMsg(self, wx_pid: int) -> int:
        ...

    def CChangeWeChatVer(self, wx_pid: int, version: str) -> int:
        ...

    def CSendAppMsg(self, wx_pid: int, wxid: str, app_id: str) -> int:
        """发送小程序"""
        ...

    def CDeleteUser(self, wx_pid: int, wxid: str) -> int:
        ...

    def CIsWxLogin(self, wx_pid: int) -> int:
        ...

    def CEditRemark(self, wx_pid: int, wxid: str, remark: str):
        ...

    def CSetChatRoomName(self, wx_pid: int, chatroomid: str, name: str):
        ...

    def CSetChatRoomAnnouncement(self, wx_pid: int, chatroomid: str, announcement: str):
        ...

    def CSetChatRoomSelfNickname(self, wx_pid: int, chatroomid: str, nickname: str):
        ...

    def CGetChatRoomMemberNickname(self, wx_pid: int, chatroomid: str, wxid: str):
        ...

    def CDelChatRoomMember(self, wx_pid: int, chatroomid: str, wxids: List[str]):
        ...

    def CAddChatRoomMember(self, wx_pid: int, chatroomid: str, wxids: List[str]):
        ...

    def COpenBrowser(self, wx_pid: int, url: str):
        """
        打开内置浏览器
        """

    def CGetHistoryPublicMsg(
        self, wx_pid: int, public_id: str, offset: str = ""
    ) -> Tuple[JsonStr]:
        """
        获取公众号历史消息

        Args:
            offset (str, optional): 起始偏移，为空的话则从新到久获取十条，该值可从返回数据中取得.
        """

    def CForwardMessage(self, wx_pid: int, wxid: str, msgid: int):
        """
        转发消息

        Args:
            wxid (str): 消息接收人
            msgid (int): 消息id
        """

    def CGetQrcodeImage(self, wx_pid: int) -> bytes:
        """
        获取二维码，同时切换到扫码登陆

        Returns:
            bytes
                二维码bytes数据.
        You can convert it to image object,like this:
        >>> from io import BytesIO
        >>> from PIL import Image
        >>> buf = wx.GetQrcodeImage()
        >>> image = Image.open(BytesIO(buf)).convert("L")
        >>> image.save('./qrcode.png')

        """

    def CGetA8Key(self, wx_pid: int, url: str) -> Tuple[JsonStr]:
        """
        获取A8Key

        Args:
            url (str): 公众号文章链接
        """

    def CSendXmlMsg(self, wx_pid: int, wxid: str, xml: str, img_path: str) -> int:
        """
        发送原始xml消息

        Returns:
            int: 0表示成功
        """

    def CLogout(self, wx_pid: int) -> int:
        """
        登出

        Returns:
            int: 0表示成功
        """

    def CGetTransfer(
        self, wx_pid: int, wxid: str, transactionid: str, transferid: int
    ) -> int:
        """
        收款

        Args:
            wxid : str
                转账人wxid.
            transcationid : str
                从转账消息xml中获取.
            transferid : str
                从转账消息xml中获取.

        Returns:
            int
                成功返回0，失败返回非0值.

        """

    def CSendEmotion(self, wx_pid: int, wxid: str, img_path: str) -> int:
        ...

    def CGetMsgCDN(self, wx_pid: int, msgid: int) -> str:
        """
        下载图片、视频、文件等

        Returns:
            str
                成功返回文件路径，失败返回空字符串.
        """


class RobotEventSinkABC:
    def OnGetMessageEvent(self, msg):
        ...


class RobotEventABC:
    def CRegisterWxPidWithCookie(self, wx_pid: int, cookie: int):
        ...



from pathlib import Path

from pydantic import BaseSettings


class Settings(BaseSettings):
    DEBUG: bool = False
    ROOT_DIR = Path(__file__).parent.parent.absolute()
    DEV_LOG_DIR = ROOT_DIR.joinpath("logs")
    DEFAULT_LOG_LEVEL = "INFO"

    class Config:
        env_file = ".env"


settings = Settings()



"""不推荐：使用Com Event"""
import json
import socket
import socketserver
import threading
import time
import typing
from collections import deque
from ctypes import Structure, c_ulonglong, c_wchar, sizeof, wintypes

import comtypes

from whochat.bot import WechatBot, WechatBotFactory
from whochat.logger import logger
from whochat.signals import Signal


class ReceiveMsgStruct(Structure):
    _fields_ = [
        ("pid", wintypes.DWORD),
        ("type", wintypes.DWORD),
        ("is_send_msg", wintypes.DWORD),
        ("msgid", c_ulonglong),
        ("sender", c_wchar * 80),
        ("wxid", c_wchar * 80),
        ("message", c_wchar * 0x1000B),
        ("filepath", c_wchar * 260),
        ("time", c_wchar * 30),
    ]

    @classmethod
    def from_bytes(cls, data: bytes) -> "ReceiveMsgStruct":
        return cls.from_buffer_copy(data)

    def to_dict(self):
        return {attname: getattr(self, attname) for attname, _ in self._fields_}


class ReceiveMsgHandler(socketserver.BaseRequestHandler):
    timeout = 3
    msg_struct_cls = ReceiveMsgStruct
    server: "WechatReceiveMsgTCPServer"
    request: socket.socket

    def setup(self) -> None:
        self.request.settimeout(self.timeout)

    def handle(self) -> None:
        comtypes.CoInitialize()
        try:
            data = self.request.recv(1024)
            struct_size = sizeof(ReceiveMsgStruct)
            while len(data) < struct_size:
                data += self.request.recv(1024)

            msg = self.msg_struct_cls.from_bytes(data)
            self._handle(msg, self.server.bot)
            self.request.sendall(b"200 OK")
        except OSError as e:
            logger.exception(e)
            return
        finally:
            comtypes.CoUninitialize()

    def _handle(self, msg: ReceiveMsgStruct, bot: typing.Optional[WechatBot] = None):
        pass


class WechatReceiveMsgTCPServer(socketserver.ThreadingTCPServer):
    """
    Com接口不支持指定IP
    """

    bot_factory = WechatBotFactory

    def __init__(
        self,
        wx_pid: int,
        port: int,
        RequestHandlerClass: typing.Callable[..., ReceiveMsgHandler],
        **kwargs,
    ):
        self.wx_pid = wx_pid
        self.port = port
        self.bot = self.bot_factory.get(wx_pid)
        super().__init__(("127.0.0.1", port), RequestHandlerClass, **kwargs)

    def serve_forever(self, poll_interval=0.5) -> None:
        logger.info(f"开始运行微信消息接收服务，地址为：{self.server_address}")
        with self.bot:
            self.bot.start_receive_message(self.port)
            super().serve_forever(poll_interval)


class StoreReceiveMsgHandler(ReceiveMsgHandler):
    def __init__(
        self, request, client_address, server, queue: deque[dict], initial_data=b""
    ):
        self.dqueue = queue
        self.initial_data = initial_data
        super(StoreReceiveMsgHandler, self).__init__(request, client_address, server)

    def handle(self) -> None:
        data = self.initial_data
        comtypes.CoInitialize()
        try:
            while True:
                data += self.request.recv(1024)
                struct_size = sizeof(ReceiveMsgStruct)
                while len(data) < struct_size:
                    buf = self.request.recv(1024)
                    if not buf:
                        return
                    data += buf
                msg = self.msg_struct_cls.from_bytes(data)
                self._handle(msg, self.server.bot)
                data = b""
                self.request.sendall(b"200 OK")
        except OSError as e:
            logger.exception(e)
        finally:
            comtypes.CoUninitialize()

    def _handle(self, msg: ReceiveMsgStruct, bot: typing.Optional[WechatBot] = None):
        msg_dict = msg.to_dict()
        logger.info(msg_dict)
        self.dqueue.append(msg_dict)


class WechatReceiveMsgRedirectTCPServer(WechatReceiveMsgTCPServer):
    RequestHandlerClass: typing.Callable[..., StoreReceiveMsgHandler]
    dqueue: deque[dict]

    def __init__(
        self,
        wx_pid: int,
        port: int,
        redirect_key: bytes,
        RequestHandlerClass: typing.Callable[
            ..., StoreReceiveMsgHandler
        ] = StoreReceiveMsgHandler,
        **kwargs,
    ):
        self.port = port
        self.dqueue = deque(maxlen=100)
        # addr tuple -> socket client which need to redirect message
        self.redirects: set[socket.socket] = set()
        self.redirect_key = redirect_key
        self.stop_redirect = False
        assert len(self.redirect_key) <= 16
        super().__init__(wx_pid, port, RequestHandlerClass, **kwargs)
        logger.info(f"转发Key为: {self.redirect_key}")
        Signal.register_sigint(self.shutdown)

    def finish_request(self, request, client_address) -> None:
        if self.redirect_key:
            flag = request.recv(len(self.redirect_key))
            if flag == self.redirect_key:
                self.redirects.add(request)
                logger.info(f"已添加一个接受消息转发的客户端：{client_address}")
            else:
                self.RequestHandlerClass(
                    request, client_address, self, self.dqueue, initial_data=flag
                )
        else:
            self.RequestHandlerClass(request, client_address, self, self.dqueue)

    def shutdown_request(self, request) -> None:
        if request in self.redirects:
            return
        super().shutdown_request(request)

    def shutdown(self) -> None:
        super().shutdown()
        self.stop_redirect = True
        self.bot_factory.kill_robot()

    def serve_forever_in_thread(self, poll_interval=0.5, daemon=True):
        def serve_in_thread():
            comtypes.CoInitialize()
            self.serve_forever(poll_interval)
            comtypes.CoUninitialize()

        t = threading.Thread(target=serve_in_thread)
        t.setDaemon(daemon)
        t.start()
        return t

    def keep_redirect(self):
        logger.info("开启转发服务")
        while not self.stop_redirect:
            if len(self.dqueue) != 0:
                msg = self.dqueue.popleft()
                data = json.dumps(msg).encode("utf-8") + b"\n"
                bad_requests = set()
                for redirect in self.redirects:
                    try:
                        redirect.sendall(data)
                    except Exception:
                        bad_requests.add(redirect)
                for bad_request in bad_requests:
                    self.redirects.remove(bad_request)
                    self.shutdown_request(bad_request)
                    logger.info("已移除一个接受消息转发的客户端")
            time.sleep(0.1)
        logger.info("转发服务已关闭")

    def serve(self):
        self.serve_forever_in_thread()
        self.keep_redirect()



import inspect
from collections import OrderedDict

from whochat.rpc.handlers import make_rpc_methods

_docs = []


def make_docs():
    if _docs:
        return _docs
    rpc_methods = make_rpc_methods()

    for name, rpc_method in OrderedDict(
        {key: rpc_methods[key] for key in sorted(rpc_methods.keys())}
    ).items():
        s = inspect.signature(rpc_method)
        description = rpc_method.__doc__
        params = []
        if rpc_method.__qualname__.split(".", maxsplit=1)[0] == "WechatBot":
            params.append(
                {
                    "name": "wx_pid",
                    "default": None,
                    "required": True,
                }
            )
        for param in s.parameters.values():
            if param.name == "self":
                continue
            params.append(
                {
                    "name": param.name,
                    "default": None if param.default is param.empty else param.default,
                    "required": param.default is param.empty,
                }
            )
        _docs.append({"name": name, "description": description, "params": params})

    return _docs


def pretty_docs():
    docs = make_docs()
    s = ""
    for item in docs:
        s += f"Name: `{item['name']}`\n"
        s += f"Description: \n\t{item['description'].lstrip() if item['description'] else '无'}\n"
        s += "Params: \n\t"
        for param in item["params"]:
            s += f"Name: `{param['name']}`\n\t"
            s += f"Required: `{str(param['required']).lower()}`\n\t"
            s += (
                f"Default: `{param['default']}`\n"
                if param["default"] is not None
                else "\n\t"
            )

        s += "\n-----------------------------------------------------\n"
    return s



import json

from fastapi import FastAPI, Request, Response
from fastapi.responses import RedirectResponse
from jsonrpcserver import async_dispatch

from whochat.rpc.docs import make_docs

app = FastAPI(title="微信机器人RPC接口文档", description="HTTP和Websocket均使用JSON-RPC2.0进行函数调用")


@app.get(
    "/",
    name="点击查看已有的RPC接口",
    description=f"```json{json.dumps(make_docs(), indent=4, ensure_ascii=False)}```",
)
async def index():
    return RedirectResponse("/docs")


@app.get("/rpc_api_docs")
async def rpc_docs():
    return make_docs()


@app.post(
    "/",
    name="RPC调用接口",
    description="""```$ curl -X POST http://localhost:5000 -d '{"jsonrpc": "2.0", "method": "start_robot_service", "params": [23472], "id": 1}'```""",
)
async def rpc(request: Request):
    data = await request.body()
    result = await async_dispatch(data)
    return Response(result)



import asyncio
import json
import logging
import re
import warnings
from collections import deque
from functools import partial
from typing import Awaitable, Callable, List

import websockets
import websockets.client
import websockets.server
from websockets.typing import Data

from whochat import _comtypes as comtypes
from whochat.abc import RobotEventSinkABC
from whochat.bot import WechatBotFactory
from whochat.signals import Signal
from whochat.utils import EventWaiter

logger = logging.getLogger("whochat")


class MessageEventStoreSink(RobotEventSinkABC):
    def __init__(self, deque_: deque):
        self.deque_ = deque_

    @staticmethod
    def _parse_extrainfo(extrainfo):
        extra = {"is_at_msg": False}
        android_windows_at_pattern = r"<atuserlist><!\[CDATA\[(.*?)\]\]></atuserlist>"
        m = re.search(android_windows_at_pattern, extrainfo)
        if not m:
            mac_at_pattern = r"<atuserlist>(.*?)</atuserlist>"
            m = re.search(mac_at_pattern, extrainfo)
        if m:
            extra["is_at_msg"] = True
            extra["at_user_list"] = m.group(1).split(",")
        m = re.search(r"<membercount>(\d+)</membercount>", extrainfo)
        if m:
            extra["member_count"] = int(m.group(1))
        return extra

    def OnGetMessageEvent(self, msg):
        if isinstance(msg, (list, tuple)):
            msg = msg[0]
        try:
            data = json.loads(msg)
            if "@chatroom" not in data["sender"]:
                data["extrainfo"] = None
            else:
                data["extrainfo"] = self._parse_extrainfo(data["extrainfo"])
        except (json.JSONDecodeError, TypeError) as e:
            logger.warning("接收消息错误: ")
            logger.exception(e)
            return
        logger.debug(f"收到消息: {data}")
        self.deque_.append(data)


class WechatMessageWebsocketServer:
    def __init__(
        self,
        wx_pids: List[int],
        ws_host: str = None,
        ws_port: int = 9001,
        queue: asyncio.Queue = None,
        welcome: bool = True,
        **kwargs,
    ):
        self.wx_pids = wx_pids
        self.ws_host = ws_host
        self.ws_port = ws_port
        self.extra_kwargs = kwargs
        self.queue = queue or asyncio.Queue(maxsize=10)
        self.welcome = welcome

        self.ws_server = None
        self.clients = set()
        self._deque = deque(maxlen=self.queue.maxsize)
        self._event_waiter = EventWaiter(2)

        self._stop_broadcast = False
        self._stop_receive_msg = False
        self._stop_websocket = asyncio.Event()
        Signal.register_sigint(self.shutdown)

    async def handler(self, websocket):
        if websocket not in self.clients:
            logger.info(f"Accept connection from {websocket.remote_address}")
            self.clients.add(websocket)
            if self.welcome:
                await websocket.send("hello")
            await websocket.wait_closed()
            self.clients.remove(websocket)
            logger.info(f"Connection from {websocket.remote_address} was closed")

    async def serve_websocket(self):
        async with websockets.server.serve(
            self.handler, self.ws_host, self.ws_port, **self.extra_kwargs
        ) as ws_server:
            logger.info(f"开始运行微信Websocket服务，地址为：<{self.ws_host}:{self.ws_port}>")
            self.ws_server = ws_server
            async with ws_server:
                await self._stop_websocket.wait()
            logger.info("Websocket服务已停止")

    def _start_receive_msg(self, wx_pids):
        comtypes.CoInitialize()
        try:
            sink = MessageEventStoreSink(self._deque)
            for wx_pid in wx_pids:
                bot = WechatBotFactory.get(wx_pid)
                bot.start_robot_service()
                logger.info(bot.get_self_info())
                logger.info("开启Robot消息推送")
                bot.start_receive_message(0)
                bot.register_event(sink)
            self._event_waiter.create_handle()
            self._event_waiter.wait_forever()
        finally:
            comtypes.CoUninitialize()

    async def deque_to_queue(self):
        while not self._stop_receive_msg:
            try:
                e = self._deque.popleft()
                await self.queue.put(e)
            except IndexError:
                await asyncio.sleep(0.1)

    async def start_receive_msg(self):
        logger.info("开始运行微信消息接收服务")
        loop = asyncio.get_event_loop()
        loop.create_task(self.deque_to_queue())
        await loop.run_in_executor(None, partial(self._start_receive_msg, self.wx_pids))
        logger.info("微信消息接收服务已停止")

    def stop_websocket(self):
        self._stop_websocket.set()

    def stop_receive_msg(self):
        self._event_waiter.stop()
        self._stop_receive_msg = True
        for wx_pid in self.wx_pids:
            bot = WechatBotFactory.get(wx_pid)
            bot.stop_receive_message()
            bot.stop_robot_service()

    def stop_broadcast(self):
        self._stop_broadcast = True
        try:
            self.queue.put_nowait("bye")
        except asyncio.QueueFull:
            pass

    async def broadcast_received_msg(self):
        logger.info("开始向客户端广播接收到的微信消息")
        while not self._stop_broadcast:
            data = await self.queue.get()
            if not self._stop_broadcast:
                self.broadcast(json.dumps(data))
        logger.info("广播已停止")

    def broadcast(self, data):
        logger.debug(f"广播消息：{data}")
        websockets.broadcast(self.clients, data)

    def shutdown(self):
        logger.info("停止服务中...")
        self.stop_websocket()
        self.stop_broadcast()
        self.stop_receive_msg()

    async def serve(self):
        try:
            websocket_task = asyncio.create_task(self.serve_websocket())
            receive_msg_task = asyncio.create_task(self.start_receive_msg())
            broadcast_task = asyncio.create_task(self.broadcast_received_msg())

            await websocket_task
            await receive_msg_task
            await broadcast_task
        except Exception as e:
            logger.exception(e)
            raise


class WechatWebsocketServer(WechatMessageWebsocketServer):
    def __init__(self, *args, **kwargs):
        warnings.warn(" 'WechatWebsocketServer' 已更名为 'WechatMessageWebsocketServer'")
        super().__init__(*args, **kwargs)


class WechatMessageWebsocketClient:
    def __init__(
        self,
        ws_uri: str,
    ):
        self.ws_uri = ws_uri

    async def start_consumer(self, on_message: Callable[[Data], Awaitable]):
        logger.info("Starting message consumer...")
        async for websocket in websockets.client.connect(self.ws_uri):
            websocket: "websockets.client.WebSocketClientProtocol"
            logger.info(f"Websocket client bind on {websocket.local_address}")
            try:
                async for message in websocket:
                    await on_message(message)
            except websockets.ConnectionClosed:
                continue



import asyncio
import dataclasses
import functools
import logging
import time
from concurrent.futures import ThreadPoolExecutor
from typing import Callable, Dict, List

import schedule
from jsonrpcserver import InvalidParams, Success

from whochat import _comtypes as comtypes
from whochat.bot import WechatBot, WechatBotFactory
from whochat.signals import Signal

logger = logging.getLogger("whochat")

bot_executor = ThreadPoolExecutor(
    max_workers=4,
    initializer=comtypes.CoInitializeEx,
    initargs=(comtypes.COINIT_APARTMENTTHREADED,),
)


class BotRpcHelper:
    bot_methods = {
        method.__name__: method
        for method in [
            WechatBot.add_brand_contact,
            WechatBot.add_friend_by_wxid,
            WechatBot.change_wechat_ver,
            WechatBot.delete_user,
            WechatBot.forward_message,
            WechatBot.get_a8_key,
            WechatBot.get_base_directory,
            WechatBot.get_chat_room_member_ids,
            WechatBot.get_chat_room_member_nickname,
            WechatBot.get_chat_room_members,
            WechatBot.get_db_handles,
            WechatBot.get_friend_list,
            WechatBot.get_history_public_msg,
            WechatBot.get_msg_cdn,
            WechatBot.get_qrcode_image,
            WechatBot.get_self_info,
            WechatBot.get_transfer,
            WechatBot.get_wechat_ver,
            WechatBot.get_wx_user_info,
            WechatBot.hook_image_msg,
            WechatBot.hook_voice_msg,
            WechatBot.is_wx_login,
            WechatBot.logout,
            WechatBot.open_browser,
            WechatBot.prevent_revoke,
            WechatBot.search_contact_by_net,
            WechatBot.send_app_msg,
            WechatBot.send_article,
            WechatBot.send_at_text,
            WechatBot.send_card,
            WechatBot.send_emotion,
            WechatBot.send_file,
            WechatBot.send_image,
            WechatBot.send_text,
            WechatBot.send_xml_msg,
            WechatBot.start_receive_message,
            WechatBot.start_robot_service,
            WechatBot.stop_receive_message,
            WechatBot.stop_robot_service,
            WechatBot.unhook_image_msg,
            WechatBot.unhook_voice_msg,
            WechatBotFactory.get_current_dir,
            WechatBotFactory.get_latest_wechat_version,
            WechatBotFactory.get_robot_pid,
            WechatBotFactory.get_we_chat_ver,
            WechatBotFactory.kill_robot,
            WechatBotFactory.list_wechat,
            WechatBotFactory.start_wechat,
        ]
    }
    rpc_methods = {}
    async_rpc_methods = {}

    @classmethod
    def make_rpc_methods(cls):
        if cls.rpc_methods:
            return cls.rpc_methods

        from whochat.bot import WechatBotFactory

        def factory(func):
            @functools.wraps(func)
            def bot_self_func(wx_pid, *args, **kwargs):
                bot = WechatBotFactory.get(wx_pid)
                _func = functools.partial(func, bot, *args, **kwargs)
                return Success(_func())

            @functools.wraps(func)
            def normal_func(*args, **kwargs):
                _func = functools.partial(func, *args, **kwargs)
                return Success(_func())

            if func.__qualname__.split(".", maxsplit=1)[0] == "WechatBot":
                return bot_self_func
            return normal_func

        for name, function in cls.bot_methods.items():
            cls.rpc_methods[name] = factory(function)
        return cls.rpc_methods

    @classmethod
    def make_async_rpc_methods(cls):
        if cls.async_rpc_methods:
            return cls.async_rpc_methods

        from whochat.bot import WechatBotFactory

        def factory(func):
            @functools.wraps(func)
            async def bot_self_func(wx_pid, *args, **kwargs):
                try:
                    bot = WechatBotFactory.get(wx_pid)
                    loop = asyncio.get_running_loop()
                    result = await loop.run_in_executor(
                        bot_executor,
                        functools.partial(func, bot, *args, **kwargs),
                    )
                    return Success(result)
                except Exception as e:
                    logger.exception(e)
                    raise

            @functools.wraps(func)
            async def normal_func(*args, **kwargs):
                try:
                    _func = functools.partial(func, *args, **kwargs)
                    loop = asyncio.get_running_loop()
                    result = await loop.run_in_executor(bot_executor, _func)
                    return Success(result)
                except Exception as e:
                    logger.exception(e)
                    raise

            if func.__qualname__.split(".", maxsplit=1)[0] == "WechatBot":
                return bot_self_func
            return normal_func

        for name, function in cls.bot_methods.items():
            cls.async_rpc_methods[name] = factory(function)
        return cls.async_rpc_methods


@dataclasses.dataclass
class BotJob:
    """

    {
        "name": "",
        "unit": "days",
        "every": 1,
        "at": "12:00:00",
        "do": {
            "func": "",
            "args": []
        },
        "description": "",
        "tags": ""
    }
    """

    name: str
    unit: str
    every: int
    at: str
    do: dict
    description: str = None
    tags: List[str] = dataclasses.field(default_factory=list)
    _job_func: Callable = dataclasses.field(init=False, default=None)

    def __post_init__(self):
        func_name, func_args = self.do["func"], self.do["args"]
        self._job_func = functools.partial(
            BotRpcHelper.make_rpc_methods()[func_name], *func_args
        )

    @property
    def job_func(self):
        return self._job_func

    def as_dict(self):
        return {
            "name": self.name,
            "unit": self.unit,
            "every": self.every,
            "at": self.at,
            "do": self.do,
            "description": self.description,
            "tags": self.tags,
        }


scheduler_executor = ThreadPoolExecutor(
    max_workers=2,
    initializer=comtypes.CoInitializeEx,
    initargs=(comtypes.COINIT_APARTMENTTHREADED,),
)


class BotScheduler:
    def __init__(self, loop=None):
        self.jobs: Dict[str, "BotJob"] = {}
        self._loop = loop
        self.executor = scheduler_executor
        self.scheduler = schedule.default_scheduler
        self.scheduled = False
        self.__shutdown = False

    @property
    def loop(self):
        if self._loop is None:
            self._loop = asyncio.get_running_loop()
        return self._loop

    def run(self):
        logger.info("开始运行任务消费线程")
        while not self.__shutdown:
            self.scheduler.run_pending()
            time.sleep(0.3)
        logger.info("任务消费线程已停止")

    def schedule(self):
        if self.scheduled:
            return
        self.scheduled = True
        Signal.register_sigint(self.shutdown)
        self.executor.submit(self.run)

    def shutdown(self):
        logger.info("正在取消所有任务...")
        self.__shutdown = True
        self._cancel_jobs()
        self.executor.shutdown(wait=False)

    async def schedule_a_job(
        self,
        name: str,
        unit: str,
        every: int,
        at: str,
        do: dict,
        description=None,
        tags=None,
    ):
        """
        {
            "name": "Greet",
            "unit": "days",
            "every": 1,
            "at": "08:00:00",
            "do": {
                "func": "send_text",
                "args": [12314, "wxid_foo", "Morning!"]
            },
            "description": "",
            "tags": ["tian"]
        }
        参见 https://schedule.readthedocs.io/en/stable/examples.html
        :param name: 任务名
        :param unit: 单位，seconds, minutes, hours, days, weeks, monday, tuesday, wednesday, thursday, friday, saturday, sunday
        :param every: 每<unit>
        :param at:  For daily jobs -> HH:MM:SS or HH:MM
                    For hourly jobs -> MM:SS or :MM
                    For minute jobs -> :SS
        :param do: 执行的方法，func: 方法名, args: 参数列表
        :param description: 描述
        :param tags: 标签，总会添加任务名作为标签
        """
        self.schedule()
        try:
            job = BotJob(name, unit, every, at, do, description, tags or [])
            assert job.name not in self.jobs
            self.jobs[job.name] = job
            schedule_job = self.scheduler.every(job.every)
            schedule_job.unit = job.unit
            schedule_job.at(job.at).do(job.job_func).tag(*job.tags, job.name)
        except AssertionError:
            return InvalidParams(f"任务<{name}>已存在")
        except Exception as e:
            return InvalidParams(f"参数错误: {str(e)}")
        return Success()

    def _cancel_jobs(self, tag=None):
        if tag is None:
            self.jobs.clear()
        self.jobs.pop(tag, None)
        self.scheduler.clear(tag=tag)
        return Success()

    async def cancel_jobs(self, tag=None):
        """
        取消任务
        :param tag: 标签名
        """
        await self.loop.run_in_executor(
            self.executor, functools.partial(self._cancel_jobs, tag=tag)
        )
        return Success()

    async def list_jobs(self):
        """列出所有任务"""
        return Success({name: job.as_dict() for name, job in self.jobs.items()})

    def get_rpc_methods(self) -> Dict[str, Callable]:
        return {
            method.__name__: method
            for method in [self.schedule_a_job, self.cancel_jobs, self.list_jobs]
        }


default_bot_scheduler = BotScheduler()


def make_rpc_methods():
    rpc_methods = BotRpcHelper.make_async_rpc_methods()
    rpc_methods.update(default_bot_scheduler.get_rpc_methods())
    return rpc_methods


def register_rpc_methods():
    from jsonrpcserver.methods import global_methods

    global_methods.update(make_rpc_methods())



import asyncio
import json
import logging
from collections import defaultdict
from functools import partial
from typing import Any, Dict

import websockets
import websockets.client
from jsonrpcclient import request as req

from whochat.rpc.handlers import make_rpc_methods

logger = logging.getLogger("whochat")

unset = object()


class Timeout(Exception):
    pass


class BotWebsocketRPCClient:
    def __init__(self, ws_uri):
        self.ws_uri = ws_uri
        self.send_queue: asyncio.Queue[Dict[str, Any]] = asyncio.Queue(maxsize=1)
        self._rpc_methods = make_rpc_methods()
        self._results = defaultdict(lambda: unset)
        self._current_request_id = None

    async def start_result_cleaner(self):
        logger.info("Starting result cleaner")
        while removable := len(self._results) - 100 > 0:
            logger.info(
                f"Current size of results: {len(self._results)}, will remove {removable} items"
            )
            for k in list(self._results.keys())[:removable]:
                self._results.pop(k, None)
            await asyncio.sleep(1)

    async def start_sender(
        self, websocket: "websockets.client.WebSocketClientProtocol"
    ):
        while not websocket.closed:
            request_dict = await self.send_queue.get()
            logger.debug(f"SEND: {request_dict}")
            try:
                await websocket.send(json.dumps(request_dict))
            except websockets.ConnectionClosed:
                await self.send_queue.put(request_dict)
                raise

    async def start_receiver(
        self, websocket: "websockets.client.WebSocketClientProtocol"
    ):
        async for message in websocket:
            logger.debug(f"RECV: {message}")
            try:
                response_dict = json.loads(message)
                if "result" in response_dict:
                    self._results[response_dict["id"]] = response_dict["result"]
                elif "error" in response_dict:
                    self._results[response_dict["id"]] = response_dict["error"]
                    logger.error(response_dict["error"])
            except json.JSONDecodeError:
                continue

    async def start_consumer(self):
        logger.info("Starting rpc client consumer")
        async for websocket in websockets.client.connect(self.ws_uri):
            websocket: "websockets.client.WebSocketClientProtocol"
            logger.info(f"Websocket client bind on {websocket.local_address}")
            gathered = asyncio.gather(
                self.start_receiver(websocket),
                self.start_sender(websocket),
            )
            try:
                await gathered
            except websockets.ConnectionClosedOK:
                logger.warning(
                    f"Websocket{websocket.local_address} closed",
                )
            except websockets.ConnectionClosedError:
                logger.warning(
                    f"Websocket{websocket.local_address} closed unexpectedly",
                )
            finally:
                gathered.cancel()

    def consume_in_background(self):
        asyncio.create_task(self.start_consumer())
        asyncio.create_task(self.start_result_cleaner())

    async def _send_and_recv(self, request):
        await self.send_queue.put(request)
        request_id = request["id"]
        while True:
            if self._results[request_id] is unset:
                await asyncio.sleep(0.1)
            else:
                return self._results[request_id]

    async def rpc_call(self, name: str, params, timeout):
        request = req(name, params)
        request_id = request["id"]
        self._current_request_id = request_id
        if timeout < 0:
            asyncio.create_task(self.send_queue.put(request))
            return request_id
        if timeout == 0:
            return await self._send_and_recv(request)
        else:
            try:
                return await asyncio.wait_for(self._send_and_recv(request), timeout)
            except asyncio.TimeoutError:
                raise Timeout(f"Timeout: rpc call timeout: {name}, params {params}")

    def __getattr__(self, item):
        def remote_func(*params, timeout=5):
            """
            :param params: 仅支持位置参数
            :param timeout: 超时时间。0: 阻塞等待返回结果, <0: 直接返回不等结果,  >0: 等待超时时间
            """
            return self.rpc_call(item, params, timeout)

        return remote_func


class OneBotWebsocketRPCClient:
    """For convenience"""

    def __init__(self, wx_pid, rpc_client: "BotWebsocketRPCClient"):
        self.wx_pid = wx_pid
        self.rpc_client = rpc_client

    def __getattr__(self, item):
        return partial(getattr(self.rpc_client, item), self.wx_pid)



import functools
import json
import os
import pathlib
import re
import shutil
import tempfile
import threading
import time
from datetime import datetime
from typing import Dict, List, Union, overload
from urllib import request

import psutil

from ._comtypes import client as com_client
from .abc import CWechatRobotABC, RobotEventABC, RobotEventSinkABC
from .logger import logger
from .utils import guess_wechat_base_directory, guess_wechat_user_by_paths

_robot_local = threading.local()


# 保证每个线程的COM对象独立
def get_robot_object(com_id) -> CWechatRobotABC:
    if not hasattr(_robot_local, "robot_object"):
        setattr(_robot_local, "robot_object", com_client.CreateObject(com_id))
    return getattr(_robot_local, "robot_object")


def get_robot_event(com_id) -> RobotEventABC:
    if not hasattr(_robot_local, "robot_event"):
        setattr(_robot_local, "robot_event", com_client.CreateObject(com_id))
    return getattr(_robot_local, "robot_event")


def auto_start(func):
    @functools.wraps(func)
    def wrapper(self: "WechatBot", *args, **kwargs):
        self.start_robot_service()
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            logger.exception(e)

    return wrapper


class WechatBot:
    def __init__(
        self,
        wx_pid,
        robot_object_id: str,
        robot_event_id: str = None,
        wechat_version=None,
    ):
        self.wx_pid = int(wx_pid)
        self._robot_object_id = robot_object_id
        self._robot_event_id = robot_event_id
        self.started = False
        self.user_info = {}
        self.event_connection = None

        self._base_directory = None
        self.image_hook_path = None
        self.voice_hook_path = None
        self.wechat_version = wechat_version

    @property
    def robot(self):
        return get_robot_object(self._robot_object_id)

    @property
    def robot_event(self):
        return get_robot_event(self._robot_event_id)

    def __enter__(self):
        self.start_robot_service()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop_robot_service()

    def start_robot_service(self):
        if self.started:
            return True
        result = self.robot.CStartRobotService(
            self.wx_pid,
        )
        if result == 0:
            self.started = True
            if self.wechat_version:
                self.change_wechat_ver(self.wechat_version)
                logger.info(f"更改微信版本号为: {self.wechat_version}")
        return not result

    def stop_robot_service(self):
        if not self.started:
            return True
        result = self.robot.CStopRobotService(
            self.wx_pid,
        )
        if result == 0:
            self.started = False
        return not result

    @property
    def base_directory(self):
        if not self._base_directory:
            return self.get_base_directory()
        return self._base_directory

    def get_base_directory(self):
        p = psutil.Process(pid=self.wx_pid)
        base_directory = guess_wechat_base_directory([f.path for f in p.open_files()])
        self._base_directory = base_directory
        return base_directory

    @overload
    def send_image(self, wx_id, img_path) -> int:
        ...

    @auto_start
    def send_image(self, wx_id, img_path):
        path = pathlib.Path(img_path)
        if not path.is_absolute():
            base_dir = self.image_hook_path or self.base_directory
            path = pathlib.Path(base_dir).joinpath(path)

        # BUG: (20220821)图片文件名如果有多个"."的话不能成功发送, 所以复制到临时文件进行发送
        if "." in path.stem:
            with tempfile.TemporaryDirectory() as td:
                temp_path = os.path.join(td, os.urandom(5).hex() + path.suffix)
                shutil.copy(path, temp_path)
                return self.robot.CSendImage(self.wx_pid, wx_id, temp_path)

        return self.robot.CSendImage(self.wx_pid, wx_id, str(path))

    @overload
    def send_text(self, wx_id, text) -> int:
        ...

    @auto_start
    def send_text(self, wx_id, text):
        return self.robot.CSendText(self.wx_pid, wx_id, text)

    @overload
    def send_file(self, wx_id, filepath) -> int:
        ...

    @auto_start
    def send_file(self, wx_id, filepath):
        path = pathlib.Path(filepath)
        if not path.is_absolute():
            path = pathlib.Path(self.base_directory).joinpath(path)

        return self.robot.CSendFile(self.wx_pid, wx_id, str(path))

    @overload
    def send_article(
        self, wxid: str, title: str, abstract: str, url: str, imgpath: str
    ):
        ...

    @auto_start
    def send_article(
        self, wxid: str, title: str, abstract: str, url: str, imgpath: str
    ):
        return self.robot.CSendArticle(self.wx_pid, wxid, title, abstract, url, imgpath)

    @overload
    def send_card(self, wxid: str, shared_wxid: str, nickname: str) -> int:
        ...

    @auto_start
    def send_card(self, wxid: str, shared_wxid: str, nickname: str):
        return self.robot.CSendCard(self.wx_pid, wxid, shared_wxid, nickname)

    @overload
    def send_at_text(
        self,
        chatroom_id: str,
        at_wxids: Union[str, List[str]],
        text: str,
        auto_nickname: bool = True,
    ) -> int:
        ...

    @auto_start
    def send_at_text(
        self,
        chatroom_id: str,
        at_wxids: Union[str, List[str]],
        text: str,
        auto_nickname: bool = True,
    ):
        return self.robot.CSendAtText(
            self.wx_pid, chatroom_id, at_wxids, text, auto_nickname
        )

    @overload
    def get_friend_list(self) -> List[Dict]:
        ...

    @auto_start
    def get_friend_list(self):
        return [
            dict(item)
            for item in (
                self.robot.CGetFriendList(
                    self.wx_pid,
                )
                or []
            )
        ]

    @overload
    def get_wx_user_info(self, wxid: str):
        ...

    @auto_start
    def get_wx_user_info(self, wxid: str):
        return json.loads(self.robot.CGetWxUserInfo(self.wx_pid, wxid))

    @property
    def wxid(self):
        return self.get_self_info()["wxId"]

    @property
    def self_path(self):
        return os.path.join(self.base_directory, str(self.wx_pid))

    @overload
    def get_self_info(self, refresh=False) -> Dict:
        ...

    @auto_start
    def get_self_info(self, refresh=False):
        if refresh or not self.user_info:
            self.user_info = json.loads(
                self.robot.CGetSelfInfo(
                    self.wx_pid,
                )
            )
        return self.user_info

    @overload
    def check_friend_status(self, wxid: str) -> int:
        ...

    @auto_start
    def check_friend_status(self, wxid: str):
        return self.robot.CCheckFriendStatus(self.wx_pid, wxid)

    @overload
    def get_com_work_path(self) -> int:
        ...

    @auto_start
    def get_com_work_path(self):
        return self.robot.CGetComWorkPath(
            self.wx_pid,
        )

    @overload
    def start_receive_message(self, port: int) -> int:
        ...

    @auto_start
    def start_receive_message(self, port: int):
        """
        开始接收消息
        :param port: 端口， port为0则使用COM Event推送
        """
        return self.robot.CStartReceiveMessage(self.wx_pid, port)

    @overload
    def stop_receive_message(self) -> int:
        ...

    @auto_start
    def stop_receive_message(self):
        return self.robot.CStopReceiveMessage(
            self.wx_pid,
        )

    @overload
    def get_chat_room_member_ids(self, chatroom_id: str) -> List:
        ...

    @auto_start
    def get_chat_room_member_ids(self, chatroom_id: str):
        wx_ids_str: str = self.robot.CGetChatRoomMembers(self.wx_pid, chatroom_id)[1][1]
        wx_ids = wx_ids_str.split("^G")
        return wx_ids

    @overload
    def get_chat_room_member_nickname(self, chatroom_id: str, wxid: str) -> str:
        ...

    @auto_start
    def get_chat_room_member_nickname(self, chatroom_id: str, wxid: str):
        return self.robot.CGetChatRoomMemberNickname(self.wx_pid, chatroom_id, wxid)

    def get_chat_room_members(self, chatroom_id: str) -> List[dict]:
        """
        获取群成员id及昵称信息

        [
            {
                "wx_id": "",
                "nickname": ""
            }
        ]
        """
        results = []
        for wx_id in self.get_chat_room_member_ids(chatroom_id):
            results.append(
                {
                    "wx_id": wx_id,
                    "nickname": self.get_chat_room_member_nickname(chatroom_id, wx_id),
                }
            )
        return results

    @overload
    def add_friend_by_wxid(self, wxid: str, message: str) -> int:
        ...

    @auto_start
    def add_friend_by_wxid(self, wxid: str, message: str):
        return self.robot.CAddFriendByWxid(self.wx_pid, wxid, message)

    @overload
    def search_contact_by_net(self, keyword: str) -> List[Dict]:
        ...

    @auto_start
    def search_contact_by_net(self, keyword: str):
        return [
            dict(item) for item in self.robot.CSearchContactByNet(self.wx_pid, keyword)
        ]

    @overload
    def add_brand_contact(self, public_id: str) -> int:
        ...

    @auto_start
    def add_brand_contact(self, public_id: str):
        return self.robot.CAddBrandContact(self.wx_pid, public_id)

    @overload
    def change_wechat_ver(self, version: str) -> int:
        ...

    @auto_start
    def change_wechat_ver(self, version: str):
        return self.robot.CChangeWeChatVer(self.wx_pid, version)

    @overload
    def send_app_msg(self, wxid: str, app_id: str) -> int:
        ...

    @auto_start
    def send_app_msg(self, wxid: str, app_id: str):
        return self.robot.CSendAppMsg(self.wx_pid, wxid, app_id)

    @overload
    def delete_user(self, wxid: str) -> int:
        ...

    @auto_start
    def delete_user(self, wxid: str):
        return self.robot.CDeleteUser(self.wx_pid, wxid)

    @overload
    def is_wx_login(self) -> int:
        ...

    @auto_start
    def is_wx_login(self) -> int:
        return self.robot.CIsWxLogin(
            self.wx_pid,
        )

    @overload
    def get_db_handles(self) -> List[Dict]:
        ...

    @auto_start
    def get_db_handles(self):
        return [dict(item) for item in self.robot.CGetDbHandles(self.wx_pid)]

    @overload
    def register_event(self, event_sink: RobotEventSinkABC) -> int:
        ...

    @auto_start
    def register_event(self, event_sink: RobotEventSinkABC):
        if self.event_connection:
            self.event_connection.__del__()
        self.event_connection = com_client.GetEvents(self.robot_event, event_sink)
        self.robot_event.CRegisterWxPidWithCookie(
            self.wx_pid, self.event_connection.cookie
        )

    @overload
    def get_wechat_ver(self) -> str:
        ...

    @auto_start
    def get_wechat_ver(self) -> str:
        return self.robot.CGetWeChatVer()

    @overload
    def hook_voice_msg(self, savepath: str) -> Union[int, str]:
        ...

    @auto_start
    def hook_voice_msg(self, savepath: str) -> Union[int, str]:
        p = pathlib.Path(savepath)
        if not p.is_absolute():
            p = pathlib.Path.home().joinpath(p)
        result = self.robot.CHookVoiceMsg(self.wx_pid, str(p))
        if result == 0:
            self.voice_hook_path = str(p)
            return self.voice_hook_path
        return result

    @overload
    def unhook_voice_msg(self) -> int:
        ...

    @auto_start
    def unhook_voice_msg(self):
        return self.robot.CUnHookVoiceMsg(self.wx_pid)

    @overload
    def hook_image_msg(self, savepath: str) -> Union[int, str]:
        ...

    @auto_start
    def hook_image_msg(self, savepath: str) -> Union[int, str]:
        p = pathlib.Path(savepath)
        if not p.is_absolute():
            p = pathlib.Path.home().joinpath(p)
        result = self.robot.CHookImageMsg(self.wx_pid, str(p))
        if result == 0:
            self.image_hook_path = str(p)
            return self.image_hook_path
        return result

    @overload
    def unhook_image_msg(self) -> int:
        ...

    @auto_start
    def unhook_image_msg(self):
        return self.robot.CUnHookImageMsg(self.wx_pid)

    @overload
    def open_browser(self, url: str) -> int:
        ...

    @auto_start
    def open_browser(self, url: str):
        return self.robot.COpenBrowser(self.wx_pid, url)

    @overload
    def get_history_public_msg(self, public_id: str, offset: str = "") -> List:
        ...

    @auto_start
    def get_history_public_msg(self, public_id: str, offset: str = ""):
        """
        获取公众号历史消息

        Args:
            offset (str, optional): 起始偏移，为空的话则从新到久获取十条，该值可从返回数据中取得
        """
        r = self.robot.CGetHistoryPublicMsg(self.wx_pid, public_id, offset)
        try:
            msgs = json.loads(r[0])
        except (IndexError, json.JSONDecodeError) as e:
            logger.exception(e)
            return []
        return msgs

    @overload
    def forward_message(self, wxid: str, msgid: int) -> int:
        ...

    @auto_start
    def forward_message(self, wxid: str, msgid: int):
        """
        转发消息

        Args:
            wxid (str): 消息接收人
            msgid (int): 消息id
        """
        return self.robot.CForwardMessage(self.wx_pid, wxid, msgid)

    @overload
    def get_qrcode_image(
        self,
    ) -> bytes:
        ...

    @auto_start
    def get_qrcode_image(
        self,
    ) -> bytes:
        """
        获取二维码，同时切换到扫码登陆

        Returns:
            bytes
                二维码bytes数据.
        You can convert it to image object,like this:
        >>> from io import BytesIO
        >>> from PIL import Image
        >>> buf = wx.GetQrcodeImage()
        >>> image = Image.open(BytesIO(buf)).convert("L")
        >>> image.save('./qrcode.png')
        """
        return self.robot.CGetQrcodeImage(self.wx_pid)

    @overload
    def get_a8_key(self, url: str) -> str:
        ...

    @auto_start
    def get_a8_key(self, url: str):
        """
        获取A8Key

        Args:
            url (str): 公众号文章链接
        """
        r = self.robot.CGetA8Key(self.wx_pid, url)
        try:
            result = json.loads(r[0])
        except (IndexError, json.JSONDecodeError) as e:
            logger.exception(e)
            return ""
        return result

    @overload
    def send_xml_msg(self, wxid: str, xml: str, img_path: str) -> int:
        ...

    @auto_start
    def send_xml_msg(self, wxid: str, xml: str, img_path: str) -> int:
        """
        发送原始xml消息

        Returns:
            int: 0表示成功
        """
        return self.robot.CSendXmlMsg(self.wx_pid, wxid, xml, img_path)

    @overload
    def logout(self) -> int:
        ...

    @auto_start
    def logout(self) -> int:
        """
        登出

        Returns:
            int: 0表示成功
        """
        return self.robot.CLogout(self.wx_pid)

    @overload
    def get_transfer(self, wxid: str, transactionid: str, transferid: int) -> int:
        ...

    @auto_start
    def get_transfer(self, wxid: str, transactionid: str, transferid: int) -> int:
        """
        收款

        Args:
            wxid : str
                转账人wxid.
            transcationid : str
                从转账消息xml中获取.
            transferid : str
                从转账消息xml中获取.

        Returns:
            int
                成功返回0，失败返回非0值.
        """
        return self.robot.CGetTransfer(self.wx_pid, wxid, transactionid, transferid)

    @overload
    def send_emotion(self, wxid: str, img_path: str) -> int:
        ...

    @auto_start
    def send_emotion(self, wxid: str, img_path: str) -> int:
        return self.robot.CSendEmotion(self.wx_pid, wxid, img_path)

    @overload
    def get_msg_cdn(self, msgid: int) -> str:
        ...

    @auto_start
    def get_msg_cdn(self, msgid: int) -> str:
        """
        下载图片、视频、文件等

        Returns:
            str
                成功返回文件路径，失败返回空字符串.
        """
        return self.robot.CGetMsgCDN(self.wx_pid, msgid)

    def prevent_revoke(self, rel_path: str, hold_time: int = 120):
        """
        防止文件被删除
        通过打开文件来阻止微信将撤回的文件删除，仅Windows可用

        :param rel_path: 相对微信数据目录的路径，如微信目录为"C:\\Users\\foo\\Documents\\WeChat Files"，
                         `rel_path`为"foo.txt", 则实际文件路径为"C:\\Users\\foo\\Documents\\WeChat Files\\foo.txt"
        :param hold_time: 持续时间，秒，默认为微信撤回时间
        """
        full_path = os.path.join(self.base_directory, rel_path)
        start = time.perf_counter()
        if not os.path.isfile(full_path):
            tmp_file_path = full_path + ".wxtmp"
            if not os.path.exists(tmp_file_path):
                logger.error(
                    f"'{full_path}' is not a file and it's tmp file {tmp_file_path} not exists"
                )
                return 1

            while not os.path.isfile(full_path):
                time.sleep(0.1)
                if time.perf_counter() > start + hold_time:
                    logger.warning("File has been revoked")
                    return 1

        def hold():
            logger.info(f"Try to open file {full_path} to prevent the deletion.")
            with open(full_path, "rb"):
                while not time.perf_counter() > start + hold_time:
                    time.sleep(1)
            logger.info(f"Opened file {full_path} closed.")

        t = threading.Thread(
            target=hold,
        )
        t.start()
        return 0


class WechatBotFactoryMetaclass(type):
    _robot: CWechatRobotABC
    _robot_object_id: str
    _robot_event_id: str

    @property
    def robot(cls):
        return get_robot_object(cls._robot_object_id)

    @property
    def robot_event(cls):
        return get_robot_event(cls._robot_event_id)

    @property
    def robot_pid(cls) -> int:
        return cls.robot.CStopRobotService(0)


class WechatBotFactory(metaclass=WechatBotFactoryMetaclass):
    """
    单例
    """

    _instances: Dict[int, "WechatBot"] = {}
    _robot_object_id = "WeChatRobot.CWeChatRobot"
    _robot_event_id = "WeChatRobot.RobotEvent"

    @classmethod
    def get(cls, wx_pid) -> "WechatBot":
        if wx_pid not in cls._instances:
            wechat_version = os.environ.get(
                "WHOCHAT_WECHAT_VERSION"
            ) or cls.get_latest_wechat_version(fill=".99")
            bot = WechatBot(
                wx_pid,
                cls._robot_object_id,
                cls._robot_event_id,
                wechat_version=wechat_version,
            )
            cls._instances[wx_pid] = bot
        return cls._instances[wx_pid]

    @classmethod
    def exit(cls):
        logger.info("卸载注入的dll...")
        for instance in cls._instances.values():
            instance.stop_robot_service()

    @classmethod
    def get_we_chat_ver(cls) -> str:
        return cls.robot.CGetWeChatVer()

    @classmethod
    def start_wechat(cls) -> int:
        return cls.robot.CStartWeChat()

    @classmethod
    def list_wechat(cls) -> List[dict]:
        results = []
        for process in psutil.process_iter():
            if process.name().lower() == "wechat.exe":
                files = [f.path for f in process.open_files()]
                results.append(
                    {
                        "pid": process.pid,
                        "started": datetime.fromtimestamp(
                            process.create_time()
                        ).isoformat(),
                        "status": process.status(),
                        "wechat_user": guess_wechat_user_by_paths(files),
                        "base_directory": guess_wechat_base_directory(files),
                    }
                )
        return results

    @classmethod
    def get_robot_pid(cls):
        return cls.robot_pid

    @classmethod
    def kill_robot(cls):
        bot_pid = cls.robot_pid
        logger.info(f"尝试结束CWeChatRobot进程: pid {bot_pid}")
        try:
            psutil.Process(bot_pid).kill()
            logger.info("OK")
        except psutil.Error as e:
            logger.warning(e, exc_info=True)

    @classmethod
    def register_events(cls, wx_pids, event_sink: RobotEventSinkABC):
        connection = com_client.GetEvents(cls.robot_event, event_sink)
        for wx_pid in wx_pids:
            cls.robot_event.CRegisterWxPidWithCookie(wx_pid, connection.cookie)

    @classmethod
    def get_current_dir(cls):
        return os.getcwd()

    @classmethod
    def get_latest_wechat_version(cls, fill: str = None):
        logger.info("获取微信最新版本号...")
        with request.urlopen(
            request.Request(
                "https://pc.weixin.qq.com/?lang=zh_CN",
                headers={
                    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/113.0.0.0 Safari/537.36"
                },
                method="GET",
            )
        ) as fp:
            html = fp.read().decode()
        m = re.search(r"download-version\">(.*?)</span>", html)
        if not m:
            logger.warning("获取微信最新版本号异常")
            return
        version = m.group(1)
        if fill:
            version = version + fill
        logger.info(f"微信最新版本号：{version}")
        return version



"""
https://github.com/ljc545w/ComWeChatRobot/releases
https://github.com/amchii/ComWeChatRobot/releases
"""
import os.path

from whochat.utils import as_admin

CWeChatRobotEXE_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "CWeChatRobot.exe")
)

__wechat_version__ = "3.7.0.30"
__robot_commit_hash__ = "f6d7577"


def register():
    as_admin(CWeChatRobotEXE_PATH, "/regserver")


def unregister():
    as_admin(CWeChatRobotEXE_PATH, "/unregserver")



import ctypes
import random
import re
import socket
import string
import sys
from typing import List, Optional


def windows_only():
    if sys.platform != "win32":
        raise RuntimeError("仅支持在win32环境下使用")


def get_free_port():
    s = socket.socket()
    s.bind(("", 0))
    ip, port = s.getsockname()
    s.close()
    return port


def get_random_string(size):
    sample = string.ascii_letters + string.digits
    if size <= len(sample):
        return "".join(random.sample(sample, k=size))
    return get_random_string(len(sample)) + get_random_string(size - len(sample))


# Windows Only
def guess_wechat_user_by_path(path: str) -> Optional[str]:
    """

    :param path: e.g. "C:\\Users\\foo\\Documents\\WeChat Files\\<wxid>\\Msg\\Sns.db"
    """

    m = re.match(r".*?WeChat Files\\([\w\-]+)\\", path, flags=re.I)
    if not m:
        return ""
    return m.group(1)


def guess_wechat_user_by_paths(paths: List[str]) -> Optional[str]:
    for path in paths:
        user = guess_wechat_user_by_path(path)
        if user:
            return user
    return ""


def guess_wechat_base_directory_by_path(path: str) -> Optional[str]:
    """
    :param path: e.g. "C:\\Users\\foo\\Documents\\WeChat Files\\<wxid>\\Msg\\Sns.db"
    """
    m = re.match(r"(.*?\\WeChat Files).*", path, flags=re.I)
    if not m:
        return ""
    return m.group(1)


def guess_wechat_base_directory(paths: List[str]) -> Optional[str]:
    for path in paths:
        directory = guess_wechat_base_directory_by_path(path)
        if directory:
            return directory
    return ""


_handles_type = ctypes.c_void_p * 1
RPC_S_CALLPENDING = -2147417835


class EventWaiter:
    def __init__(self, timeout: int):
        self.timeout = timeout
        self.handle = None
        self._handles = None

    def reset_event(self):
        return ctypes.windll.kernel32.ResetEvent(self.handle)

    def set_event(self):
        return ctypes.windll.kernel32.SetEvent(self.handle)

    def create_handle(self):
        self.handle = ctypes.windll.kernel32.CreateEventA(None, True, False, None)
        self._handles = _handles_type(self.handle)

    def close_handle(self):
        return ctypes.windll.kernel32.CloseHandle(self.handle)

    def stop(self):
        return self.set_event() != 0

    def wait_once(self) -> bool:
        try:
            res = ctypes.oledll.ole32.CoWaitForMultipleHandles(
                0,
                int(self.timeout * 1000),
                len(self._handles),
                self._handles,
                ctypes.byref(ctypes.c_ulong()),
            )
            return res == 0
        except WindowsError as details:
            if details.winerror == RPC_S_CALLPENDING:  # timeout expired
                return False
            raise

    def wait_forever(self):
        self.reset_event()
        try:
            while not self.wait_once():
                pass
        finally:
            self.close_handle()


def as_admin(exe: str, params: str = None):
    ctypes.windll.shell32.ShellExecuteW(None, "runas", exe, params)



import logging
import os
from logging.handlers import RotatingFileHandler

import click

from whochat.rpc.handlers import register_rpc_methods
from whochat.utils import windows_only


@click.group(
    name="whochat",
)
@click.option(
    "--log-level",
    "-l",
    "log_level",
    default="info",
    show_default=True,
    help="日志等级, `debug`, `info`, `warn`, `error`",
)
@click.option("--log-file", "log_file", help="日志文件, 可以是相对路径")
def whochat(log_level: str, log_file):
    """
    微信机器人

    使用<子命令> --help查看使用说明
    """
    logger = logging.getLogger("whochat")
    logger.setLevel(log_level.upper())
    if log_file:
        abs_log_file = os.path.abspath(log_file)
        dirname = os.path.dirname(abs_log_file)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
        assert os.path.isdir(dirname), f"{dirname}不存在或不是一个目录"
        file_handler = RotatingFileHandler(
            abs_log_file,
            maxBytes=1024 * 1024,  # 1MB
            backupCount=10,
            encoding="utf-8",
        )
        verbose_formatter = logging.Formatter(
            "[%(levelname)s] [%(name)s] %(asctime)s %(filename)s %(process)d %(message)s"
        )
        file_handler.setFormatter(verbose_formatter)
        logger.addHandler(file_handler)


@whochat.command()
def version():
    """显示程序和支持微信的版本信息"""
    from whochat import __version__
    from whochat.ComWeChatRobot import __robot_commit_hash__, __wechat_version__

    s = ""
    s += f"WhoChat version: {__version__}\n"
    s += f"Supported WeChat version: {__wechat_version__}\n"
    s += f"Build ComWechatRobot's dll from: {__robot_commit_hash__}"
    click.echo(s)


@whochat.command()
@click.option("--unreg", "-R", is_flag=True, default=False, help="取消注册")
def regserver(unreg):
    """注册COM"""
    windows_only()
    from whochat.ComWeChatRobot import register, unregister

    if unreg:
        unregister()
    else:
        register()


@whochat.command()
def list_wechat():
    """列出当前运行的微信进程"""
    windows_only()

    from whochat.bot import WechatBotFactory

    s = ""
    for r in WechatBotFactory.list_wechat():
        s += f"PID: {r['pid']}" + "\n"
        s += f"启动时间: {r['started']}" + "\n"
        s += f"运行状态: {r['status']}" + "\n"
        s += f"用户名: {r['wechat_user']}" + "\n"
        s += "---\n"
    click.echo(s)


@whochat.command()
@click.option(
    "--host", "-h", default="localhost", show_default=True, help="Server host."
)
@click.option("--port", "-p", default=9001, show_default=True, help="Server port")
@click.option(
    "--welcome",
    default=True,
    type=bool,
    show_default=True,
    help="Respond 'hello' on client connect",
)
@click.argument("wx_pids", nargs=-1, type=int)
def serve_message_ws(host, port, welcome, wx_pids):
    """
    运行接收微信消息的Websocket服务

    WX_PIDS: 微信进程PID
    """
    windows_only()

    import asyncio

    from whochat.messages.websocket import WechatMessageWebsocketServer

    if not wx_pids:
        raise click.BadArgumentUsage("请指定至少一个微信进程PID")

    async def main():
        server = WechatMessageWebsocketServer(
            wx_pids=wx_pids, ws_host=host, ws_port=port, welcome=welcome
        )
        await server.serve()

    asyncio.run(main())


@whochat.command()
@click.option("--json", "json_", is_flag=True, default=False, help="JSON文档")
def show_rpc_docs(json_):
    """
    列出RPC接口

    \b
    whochat show-rpc-docs
    or
    whochat show-rpc-docs --json > api.json
    """
    import json

    from whochat.rpc.docs import make_docs, pretty_docs

    if json_:
        click.echo(json.dumps(make_docs(), ensure_ascii=False, indent=4))
    else:
        click.echo(pretty_docs())


@whochat.command()
@click.option(
    "--host", "-h", default="localhost", show_default=True, help="Server host."
)
@click.option("--port", "-p", default=9002, show_default=True, help="Server port")
def serve_rpc_ws(host, port):
    """
    运行微信机器人RPC服务(JSON-RPC2.0), 使用Websocket
    """
    windows_only()

    import asyncio

    from whochat.rpc.servers.websocket import run

    click.echo(f"PID: {os.getpid()}")
    register_rpc_methods()

    asyncio.run(run(host, port))


@whochat.command()
@click.option(
    "--host", "-h", default="localhost", show_default=True, help="Server host."
)
@click.option("--port", "-p", default=9002, show_default=True, help="Server port")
def serve_rpc_http(host, port):
    """
    运行微信机器人RPC服务(JSON-RPC2.0), 使用HTTP接口

    需要安装fastapi和uvicorn: `pip install fastapi uvicorn`
    """
    windows_only()

    import uvicorn

    from whochat.rpc.servers.http import app

    register_rpc_methods()
    uvicorn.run(app, host=host, port=port)



import asyncio
import logging

import websockets.server
from jsonrpcserver import async_dispatch

from whochat.signals import Signal

logger = logging.getLogger("whochat")


async def dispatch_in_task(websocket, request):
    res = await async_dispatch(request)
    await websocket.send(res)


async def handler(websocket: "websockets.server.WebSocketServerProtocol"):
    logger.info(f"Accept connection from {websocket.remote_address}")
    while not websocket.closed:
        request = await websocket.recv()
        asyncio.create_task(dispatch_in_task(websocket, request))

    logger.info(f"Connection from {websocket.remote_address} was closed")


async def run(host, port):
    stop_event = asyncio.Event()

    def shutdown():
        logger.info("正在停止微信机器人RPC websocket服务...")
        stop_event.set()
        from whochat.bot import WechatBotFactory

        WechatBotFactory.exit()

    Signal.register_sigint(shutdown)
    async with websockets.server.serve(handler, host, port):
        logger.info(f"运行微信机器人RPC websocket服务, 地址为<{host}:{port}>")
        await stop_event.wait()



import logging
from logging.handlers import RotatingFileHandler

from whochat.settings import settings

verbose_formatter = logging.Formatter(
    "[%(levelname)s] [%(name)s] %(asctime)s %(filename)s %(process)d %(message)s"
)

logger = logging.getLogger("whochat")
logger.setLevel(settings.DEFAULT_LOG_LEVEL)
console_handler = logging.StreamHandler()
console_handler.setFormatter(verbose_formatter)
logger.addHandler(console_handler)

if settings.DEBUG:
    logger.setLevel(logging.DEBUG)
    file_handler = RotatingFileHandler(
        settings.DEV_LOG_DIR.joinpath("whochat.log"),
        maxBytes=1024 * 1024,  # 1MB
        backupCount=10,
        encoding="utf-8",
    )
    file_handler.setFormatter(verbose_formatter)
    logger.addHandler(file_handler)



import enum


class WechatMsgType(enum.IntEnum):
    文字 = 1
    图片 = 3
    语音 = 34
    好友确认 = 37
    名片 = 42
    视频 = 43
    表情 = 47
    位置 = 48
    共享实时位置_文件_转账_链接 = 49
    语音通话消息 = 50
    微信初始化 = 51
    语音通话通知 = 52
    语音通话邀请 = 53
    小视频 = 62
    SYS_NOTICE = 9999
    系统消息_红包 = 10000
    撤回_群语音邀请 = 10002



import sys

from whochat.utils import windows_only

if sys.platform == "win32":
    from comtypes import *  # noqa
    from comtypes import client  # noqa
else:

    COINIT_MULTITHREADED = 0x0
    COINIT_APARTMENTTHREADED = 0x2
    COINIT_DISABLE_OLE1DDE = 0x4
    COINIT_SPEED_OVER_MEMORY = 0x8

    class Unusable:
        def __call__(self, *args, **kwargs):
            windows_only()

    class UnusableModule:
        def __getattr__(self, item):
            return Unusable()

    CoInitialize = Unusable()
    CoUninitialize = Unusable()
    CoInitializeEx = Unusable()
    client = UnusableModule()



import signal
from collections import defaultdict

from .logger import logger


class Signal:
    _signal_handlers = defaultdict(set)
    signalled = False

    @classmethod
    def register(cls, signum, func):
        if not cls.signalled:
            cls.signal()
        assert callable(func)
        cls._signal_handlers[signum].add(func)

    @classmethod
    def register_sigint(cls, func):
        logger.info(f"注册SIGINT信号处理程序: {func.__qualname__}")
        cls.register(signal.SIGINT, func)

    @classmethod
    def unregister(cls, signum, func):
        cls._signal_handlers[signum].remove(func)

    @classmethod
    def handler(cls, signum, frame):
        logger.info(f"接收到信号: {signal.strsignal(signum)}")
        for _handler in cls._signal_handlers[signum]:
            try:
                _handler()
            except Exception as e:
                logger.error(f"信号处理程序出现错误: {_handler}")
                logger.exception(e)

    @classmethod
    def signal(cls):
        signal.signal(signal.SIGINT, cls.handler)
        cls.signalled = True
